/* ------------------------------------------------------------------
 * 
 *  main.c   Example of using Digilent PMOD Oled with freeRTOS
 * 
 *  Uses LEDs and Buttons built into MX4cK
 *  Flashes  LED #1 every 0.25 sec
 *  Flashes  LED #2 every 1 sec
 *  Button #2 counts up
 *  Button #1 counts down
 *  OLED display is update to show count
 *  A queue is used to communicate between threads
 *  Display is turned off if the user presses both buttons simultaneously.  Must restart program to
 *  reinitialize the display
 *
 *  NOTE: PMOLED library was modified for compatibility with freeRTOS
 *  Mason, Seattle U, 2013
 *
 *  Update:  4/9/2014  removed repeat feature,
 *                     added simultaneous button press for display off
 *  See freeRTOS readme
 * 
 * --------------------------------------------------------------- */



#include "MX4cKsetup.h"
#include "MX4cKsettings.h"


// the RTOS is configured with a 1ms tick length
#include "RTOS/FreeRTOS.h"
#include "RTOS/task.h"
#include "RTOS/queue.h"
#include "RTOS/timers.h"
#include "RTOS/semphr.h"

#include "PmodOLED/PmodOLED.h"


// forward declaration
static void Timer1(xTimerHandle xTimer);
static void Timer2(xTimerHandle xTimer);
static void Task1(void *pvParameters);
static void Task2(void *pvParameters);

static void SetupHardware(void);

// queue to hold the button press count
xQueueHandle countqueue;



// create user defined characters for OLED
//#define USERDEFINEDCHAR   // uncomment this line to generate userdefined characters
// see Pmod OLED Example code by DigilentInc for details
#if defined (USERDEFINEDCHAR)

#define cbPageMax	256
extern BYTE rgbOledBmp[];
BYTE rgbBmp0[] = {
    0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81,
    0x18, 0x24, 0x42, 0x81, 0x81, 0x42, 0x24, 0x18
};
BYTE rgbBmp1[32];
BYTE rgbUserFont[cbOledFontUser] = {
    0x7E, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x7E, // 0x00
    0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81, // 0x01
    0x7E, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, // 0x02
    0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7E, // 0x03
    0x7E, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, // 0x04
    0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x7E, // 0x05
    0x7E, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7E, // 0x06
    0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, // 0x07
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, // 0x08
    0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, // 0x09
    0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, // 0x0A
    0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, // 0x0B
    0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, // 0x0C
    0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, // 0x0D
    0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, // 0x0E
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, // 0x0F
    0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // 0x10
    0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // 0x11
    0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, // 0x12
    0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, 0x00, // 0x13
    0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00, // 0x14
    0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, // 0x15
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, // 0x16
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF // 0x17
};
BYTE rgbCmd[cbPageMax];
BYTE rgbBuf[cbPageMax];
#endif  // end of user defined characters code

int main(void) {
    /* Prepare the hardware to run */
    SetupHardware();

    // initialize and clear the OLED
    // must do this before initializing freeRTOS due to timer usage conflict
    OledInit();
    OledClear();

#if defined (USERDEFINEDCHAR)
    char ch;
    // copy to the userdefined character array
    for (ch = 0; ch < 0x18; ch++) {
        OledDefUserChar(ch, &rgbUserFont[ch * cbOledChar]);
    }
#endif

    // set the digital io ports
    // configure the built-in LEDs
    //   LED1 is on RB10
    //	 LED2 is on RB11
    //	 LED3 is on RB12
    //	 LED4 is on RB13
    //Set ports for onboard LEDs to outputs
    PORTSetPinsDigitalOut(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13);
    PORTClearBits(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13); // turn off the LEDs

    // set the built-in buttons as inputs
    PORTSetPinsDigitalIn(IOPORT_A, BIT_6 | BIT_7);

    // Create two tasks of equal priority.  The RTOS gives these tasks equal CPU time
    // if one of the tasks was higher priority you would need a way to release CPU time so the lower priority
    // task can run.  This can be done with a vTaskDelay(), vTaskDelayUntil(), or by yielding the task  portYIELD()

    // create a thread to update the OLED display
    xTaskCreate(Task1, // The function that implements the task.
            (signed char *) "Task1", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE, // The size of the stack to allocate to the task.
            // will need to increase this if your task uses much stack space
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            NULL); // The task handle is not required, so NULL is passed.

    // create a thread to poll and respond to buttons
    xTaskCreate(Task2, // The function that implements the task.
            (signed char *) "Task2", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE, // The size of the stack to allocate to the task.
            // will need to increase this if your task uses much stack space
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            NULL); // The task handle is not required, so NULL is passed.



    // example of creating a timer thread
    // timer functions are called periodically automatically by the RTOS
    // they automatically yeild CPU time when not in use
    xTimerHandle xTimer1;
    xTimerHandle xTimer2;

    xTimer1 = xTimerCreate((const signed char *) "Blinky", // A text name, purely to help debugging.
            250, // The timer period. 500 ticks = .25 seconds */
            pdTRUE, // This is an auto-reload timer, so xAutoReload is set to pdTRUE.
            (void *) 0, // The ID is not used, so can be set to anything.
            Timer1 // The callback function that inspects the status of all the other tasks.
            );

    xTimer2 = xTimerCreate((const signed char *) "Blinky2", // A text name, purely to help debugging.
            (1000), // The timer period.  1000 ticks = 1 second
            pdTRUE, // This is an auto-reload timer, so xAutoReload is set to pdTRUE.
            (void *) 0, //The ID is not used, so can be set to anything.
            Timer2 // The callback function that inspects the status of all the other tasks.
            );


    if (xTimer2 != NULL) {
        xTimerStart(xTimer2, 0);
    }
    if (xTimer1 != NULL) {
        xTimerStart(xTimer1, 0);
    }

    // queue to keep track of button presses
    // each slot hold the current count
    countqueue = xQueueCreate(30, 4);

    // Start the tasks and timer running. 
    vTaskStartScheduler();

    for (;;);
}

/*-----------------------------------------------------------*/

static void Timer1(xTimerHandle xTimer) {
    // This function is called when the timer expires.
    // This function toggles a LED.
    PORTToggleBits(IOPORT_B, BIT_10);
}

static void Timer2(xTimerHandle xTimer) {
    // This function is called when the  timer2 expires.
    // The function does is toggle the LED.   */
    PORTToggleBits(IOPORT_B, BIT_11);
}

// update the OLED whenever there is a change in the button press count

static void Task1(void *pvParameters) {
    // Once called the task loop runs indefinitely only
    // being interrupted by equal or higher priority tasks

    // useful functions to control timing in a task or yielding CPU time
    // portTickType xTaskGetTickCount();             // return current system tick count
    // vTaskDelay( portTickType xTicksToDelay );     // block for fix number of ticks
    // vTaskDelayUntil( portTickType *pxPreviousWakeTime,  portTickType xTimeIncrement );
    //      block until a specific time.  Use this for repeating tasks when you need to compensate for computational delay

    int currentcount = 0; // button count
    char text[20]; //  used to generate the display text
    char inttext[5];


    //   unsigned int uxHighWaterMark;  // Optional: used to track mem usage
    while (!PORTReadBits(IOPORT_A, BIT_7)  || !PORTReadBits(IOPORT_A, BIT_6)) {
        // if user pressed both button down, shut down the OLED
        // infinite loop for the task
        // do what ever you need to here


        // wait for a change in the count
        xQueueReceive(countqueue, &currentcount, portMAX_DELAY);


        OledClearBuffer();

        // display count as text
        OledSetCharUpdate(0);
        OledSetCursor(0, 0);
        strcpy(text, "count=");
        strcat(text, itoa(inttext, currentcount, 10));
        OledPutString(text);

        //  alternate text to track task memory use
        //  reports remaining number if 4 byte int remaining in stack
        //        uxHighWaterMark = uxTaskGetStackHighWaterMark( NULL );
        //        strcpy(text, "  mem:");
        //        strcat(text, itoa(inttext, uxHighWaterMark, 10));
        //        OledPutString(text);

        // horizontal line
        OledMoveTo(0, 12);
        OledLineTo(127, 12);


        // rectangular bar proportional to count
        OledMoveTo(64, 15);
        // if negative or positive count
        if (currentcount >= 0) {
            OledSetFillPattern(OledGetStdPattern(1));
        } else {
            OledSetFillPattern(OledGetStdPattern(2));
        }
        OledFillRect(currentcount + 64, 28);

        // horizontal line
        OledMoveTo(0, 31);
        OledLineTo(127, 31);
        OledUpdate();

    }
    // turn off display and do nothing
    OledClearBuffer();

    // display count as text
    OledSetCharUpdate(0);
    OledSetCursor(0, 0);

    OledPutString("Display Off");
    OledUpdate();
    vTaskDelay(3000);
    OledShutdown();

    while (1) {
        // do nothing but clean out the queue
        xQueueReceive(countqueue, &currentcount, portMAX_DELAY);
    }
}

// monitor button states
// update the queue when buttons are pressed

static void Task2(void *pvParameters) {

    int clickcount = 0; // current click count

    int currstate1 = 0; // state of two buttons
    int prevstate1 = 0;
    int currstate2 = 0;
    int prevstate2 = 0;

    int repeat = 0; // handles autorepeat when button is  held down

    xQueueSendToBack(countqueue, &clickcount, 0); // initialize the queue with something so the display will be updated
    for (;;) {

        // button #2
        currstate1 = PORTReadBits(IOPORT_A, BIT_7);

        // button changed state and is down
        if (currstate1 != prevstate1 && currstate1 != 0) {
            clickcount++;
            xQueueSendToBack(countqueue, &clickcount, 0);
            repeat = 0;
        } 

        // button #1

        currstate2 = PORTReadBits(IOPORT_A, BIT_6);

         // button changed state and is down
        if (currstate2 != prevstate2 && currstate2 != 0) {
            clickcount--;
            xQueueSendToBack(countqueue, &clickcount, 0);
            repeat = 0;
        } 

        // update past button states for next loop
        // used to track change in button state without multiple button reads
        prevstate1 = currstate1;
        prevstate2 = currstate2;

        vTaskDelay(20);
    }
}

static void SetupHardware(void) {
    /* Set the system and peripheral bus speeds and enable the program cache*/
    SYSTEMConfigPerformance(SYSCLOCK);
    mOSCSetPBDIV(OSC_PB_DIV_8);

    /* Setup to use the external interrupt controller. */
    INTEnableSystemMultiVectoredInt();

    portDISABLE_INTERRUPTS();
}

/*-----------------------------------------------------------
 *  Hooks for RTOS.  Are not required to modify these
 *  May want to turn on LED or some other external signal if an error
 *  occurs
 *-----------------------------------------------------------*/


void vApplicationMallocFailedHook(void) {
    /* vApplicationMallocFailedHook() will only be called if
    configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
    function that will get called if a call to pvPortMalloc() fails.
    pvPortMalloc() is called internally by the kernel whenever a task, queue,
    timer or semaphore is created.  It is also called by various parts of the
    demo application.  If heap_1.c or heap_2.c are used, then the size of the
    heap available to pvPortMalloc() is defined by configTOTAL_HEAP_SIZE in
    FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
    to query the size of free heap space that remains (although it does not
    provide information on how the remaining heap might be fragmented). */
    taskDISABLE_INTERRUPTS();
    for (;;);
}

/*-----------------------------------------------------------*/

void vApplicationIdleHook(void) {
    /* vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
    to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
    task.  It is essential that code added to this hook function never attempts
    to block in any way (for example, call xQueueReceive() with a block time
    specified, or call vTaskDelay()).  If the application makes use of the
    vTaskDelete() API function (as this demo application does) then it is also
    important that vApplicationIdleHook() is permitted to return to its calling
    function, because it is the responsibility of the idle task to clean up
    memory allocated by the kernel to any task that has since been deleted. */
}

/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook(xTaskHandle pxTask, signed char *pcTaskName) {
    (void) pcTaskName;
    (void) pxTask;

    /* Run time task stack overflow checking is performed if
    configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook	function is
    called if a task stack overflow is detected.  Note the system/interrupt
    stack is not checked. */
    taskDISABLE_INTERRUPTS();
    for (;;);
}

/*-----------------------------------------------------------*/

void vApplicationTickHook(void) {
    /* This function will be called by each tick interrupt if
    configUSE_TICK_HOOK is set to 1 in FreeRTOSConfig.h.  User code can be
    added here, but the tick hook is called from an interrupt context, so
    code must not attempt to block, and only the interrupt safe FreeRTOS API
    functions can be used (those that end in FromISR()). */
}

/*-----------------------------------------------------------*/

void _general_exception_handler(unsigned long ulCause, unsigned long ulStatus) {
    /* This overrides the definition provided by the kernel.  Other exceptions
    should be handled here. */
    for (;;);
}
/*-----------------------------------------------------------*/

